#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import json
import pickle
import sqlite3
import logging
logging.getLogger().setLevel(logging.INFO)

ENTITIES = ['&nbsp;', '&lt;', '&gt;', '&amp;', '&quot;',	
'&apos;', '&cent;', '&pound;', '&yen;' '&euro;', '&copy;', '&reg;']

IGNORE_SECTION = ['reference', 'references', 'see also', 'completed',
                  'track listing', 'gallery', '<gallery>']

IGNORE_THE_EXTRA_CONTENT = ['[[File:', '[[Category:', '[[Image:', '[[Category:']

#Checks the object if it's picklable or not. (:_:DEBUGGING PURPOSE:_:)
def is_picklable(obj):
    try:
        pickle.dumps(obj)
        return True
    except (pickle.PicklingError, TypeError):
        return False
    
def table_exists(cursor, table_name):
    cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name=?", (table_name,))
    return cursor.fetchone() is not None    
    
def update_sqlite_table_with_dict(database_file, table_name, data_dict):

    try:

        # Connect to the SQLite database
        conn = sqlite3.connect(database_file)
        cursor = conn.cursor()

        # Check if the table exists
        if not table_exists(cursor, table_name):
            # If the table doesn't exist, create it
            create_table_sql = f"CREATE TABLE {table_name} (key TEXT PRIMARY KEY, value TEXT);"
            cursor.execute(create_table_sql)
        
        # Iterate through the dictionary and update the database
        for key, value_list in data_dict.items():
            # Serialize the list to JSON before inserting
            value_json = json.dumps(value_list)

            # Define the SQL update statement
            update_sql = f"INSERT OR REPLACE INTO {table_name} (key, value) VALUES (?, ?)"
            
            # Execute the SQL statement with the current key and serialized value
            cursor.execute(update_sql, (key, value_json))
        # Commit the changes to the database
        conn.commit()

        # Close the cursor and connection
        cursor.close()
        conn.close()
    
    except sqlite3.Error as e:
        print(f"SQLite error: {e}")
    except Exception as e:
        print(f"Error: {e}")      

def retrieve_data_from_sqlite(database_file, table_name, chunk_size, random=False):
    data_dict = {}
    try:
        # Connect to the SQLite database.
        conn = sqlite3.connect(database_file)
        cursor = conn.cursor()

        # Retrieve data from the database.
        if random:
            cursor.execute(f"SELECT rowid FROM {table_name} ORDER BY RANDOM() LIMIT ?", (chunk_size, ))
            random_row_ids = [row[0] for row in cursor.fetchall()]
            cursor.execute("SELECT key, value FROM {} WHERE rowid IN ({})".format(table_name, ','.join(map(str, random_row_ids))))
        else:
            cursor.execute(f"SELECT key, value FROM {table_name} LIMIT ?", (chunk_size, ))  

        for key, value_json in cursor.fetchall():
            # Deserialize the JSON back to a list
            value_list = json.loads(value_json)
            data_dict[key] = value_list
    
    except sqlite3.Error as e:
        print(f"SQLite error: {e}")
    except Exception as e:
        print(f"Error: {e}")
    finally:
        # Close the cursor and connection in the finally block.
        if cursor:
            cursor.close()   
        if conn:
            conn.close()   
    return data_dict     

def retrieve_page_number(database_file, table_name):
    # Connect to your SQLite database
    connection = sqlite3.connect(f'{database_file}')

    # Create a cursor
    cursor = connection.cursor()

    # Define your SQL query with the COUNT function
    query = f"SELECT COUNT(*) FROM {table_name}"

    # Execute the query and fetch the result
    cursor.execute(query)
    count = cursor.fetchone()[0]

    # Close the cursor and the database connection
    cursor.close()
    connection.close()

    # Print the count
    return count


def get_first_sentence(text):

    #This function is for getting the first sentence of the text.
    #In this way, we can create a small size index. (Ideal)
    #The column will be used during the faiss-index creation.

    """
    :params text (str): Text to be stripped.
    """


    fs_regex = re.compile(r'\.\s[A-Z].+', flags=re.MULTILINE)
    fixed_text = re.sub(fs_regex, '', text) + '.'
    fs_check = re.compile(r'\.\s+,\s*[a-z]')
    checked = re.search(fs_check, fixed_text)
    if checked:
        part = checked.group(0)
        replacement = '. ' + part[-1].upper()
        fixed_text = fixed_text.replace(part, replacement)
    return fixed_text    

def remove_nested_curly(text):
            stringus = ''
            bumble = 0
            for i in text:
                if i == '{':
                    bumble += 1
                elif i == '}'and bumble > 0:
                    bumble -= 1
                elif bumble == 0:
                    stringus += i
            return stringus

def cleaning_text(list_of_dicts, first_sentence=False):

    """
    :params list_of_dicts: List of dicts to be cleaned.
    :params first_sentence: If True, only the first sentence of the text will be returned.
    """

    try:

        plain_text = {}

        titles, ids, texts = ([] for x in range(3))

        titles = [key for d in list_of_dicts for key in d.keys()]
        ids = [value[0] for d in list_of_dicts for value in d.values()]
        texts = [value[1] for d in list_of_dicts for value in d.values()]


        for title, id, text in zip(titles, ids, texts):

            text = re.sub("<!-+(.*?)-+>", '', text, flags=re.DOTALL)
            #Remove the part after 'IGNORE_SECTION'.
            _id = [text.find(part) for part in re.findall(r'=+(.*?)=+', text) if part.lower().strip() in IGNORE_SECTION]
            text = text[:min(_id)] if len(_id) != 0 else text

            # ==CURLY BRACKET REMOVER== # 
            text = remove_nested_curly(text)
            
            #Remove HTML-like tag pairs.
            text = re.sub("<(.*?)>(.*?)<\/(.*?)>", '', text, flags=re.MULTILINE)
            text = re.sub(r'<(.*?)>', '', text)
            
            #Checks the page and removes extra contents such as files and categories.
            for extra in IGNORE_THE_EXTRA_CONTENT:
                _ = 0
                _start = []
                _end = []
                while _ < len(text):
                    index = text.find(extra, _)
                    if index == -1:
                        break
                    _start.append(index)
                    _ = index + 1
                for idx in _start:
                    idx += 2
                    open_bracket_count = 2
                    for _letter in text[idx:]:
                        if open_bracket_count != 0:
                            if _letter == ']':
                                open_bracket_count -= 1
                            elif _letter == '[':
                                open_bracket_count += 1
                            else:
                                pass
                            idx += 1
                        else:
                            break
                    _end.append(idx)
                #Reversed lists to not cause errors during indexing.    
                reversed_start = _start[::-1]
                reversed_end = _end[::-1]
                for start, end in zip(reversed_start, reversed_end):
                    sub = text[start:end]
                    text = text.replace(sub, '')
            
            #We're accessing the data between the double brackets and check if it's useful for us or not. (Alias Remover)
            for part in re.findall(r'\[\[.+?\]\]', text, flags=re.DOTALL):
                if '|' in part:
                    second = part.split('|')[1].strip(']')
                    text = text.replace(part, second)
            
            #Then we're gonna remove all the links in the text.
            text = re.sub(r'https?:\/\/.*[\r\n]*', '', text, flags=re.MULTILINE)
            
            #Remove headlines, we don't need them.
            text = re.sub(r'=+(.*?)=+', '', text)
            #Remove the special characters, not cool.    
            for entity in ENTITIES:
                special = re.compile(entity)
                text = re.sub(special, '', text)
            #Remove the symbols we don't need.
            symbols = ["'", "{", "}", "*", '#', '"', "[", "]"]

            text = text.translate({ord(ch):'' for ch in symbols})
            text = text.replace('()', '')
            #Some additional regex for wikipedia format.
            text = re.sub(r'\n\|.*\n', "", text)
            text = re.sub(r'\|+(.*?)\|+', "", text)
            three = re.escape('align=')+r'.*\n'
            text = re.sub(three, "", text)
            
            #New line and space remover, the data must be plain text.
            text = re.sub(r'\n+', ' ', text)
            text = re.sub(r'\s\s+', ' ', text)
            #Removing the meaningless tags.
            text = text.replace('&ndash;', '-')
            
            #Removing the trailing spaces, if there is any.
            text = text.lstrip(' ').rstrip(' ')

            if first_sentence:

                text = get_first_sentence(text)

            plain_text.update({title:[id,text]})
   
        return plain_text   
     
    except Exception as e:
        print(e) 


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#     =======================================================================
# 
#     Copyright (c) 2023  @sharkcommits (sharkcommits@protonmail.com)
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#     INSTALLATION: 'pip install extractipedia'
#
#     USAGE: 'python -m extractipedia.CheckDatabase -f YOUR_DATABASE.db -t YOUR_TABLE'
#
#     HELP: 'python -m extractipedia.CheckDatabase --help' to get more information.
#
#     =======================================================================


"""

This script prints the first n items from the database like the following section:

('title_1', [id_1, 'text_1'])
('title_2', [id_2, 'text_2'])
('title_n', [id_n, 'text_n'])

"""
import gc
import argparse
from extractipedia.utils import retrieve_data_from_sqlite, retrieve_page_number

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Reading Database')
    parser.add_argument("-d", "--database_file", type=str, default='new_database.db')
    parser.add_argument("-t", "--table_name", type=str, default='new_table')
    parser.add_argument("-c", "--chunk_size", type=int, default=10, help='Retrieve n pages from the database.')
    parser.add_argument("-r", "--random", action="store_true", default=False, help='Selects random n pages. (default=False)')
    args = parser.parse_args()

    try:

        result = retrieve_data_from_sqlite(args.database_file, args.table_name, args.chunk_size, args.random)
        total_pages = retrieve_page_number(args.database_file, args.table_name)
        
        for key, value in result.items():
            print('='*10 + key + '='*10, value, sep='\n')
        print(f"\nYou have successfully retrieved {args.chunk_size} PAGES, you've got TOTAL {total_pages} PAGES.\n")    

        del result
        _ = gc.collect()    

    except Exception as e:

        print(e)        


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#     =======================================================================
# 
#     Copyright (c) 2023  @sharkcommits (sharkcommits@protonmail.com)
# 
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#     INSTALLATION: 'pip install extractipedia'
#
#     USAGE: 'python -m extractipedia.Extraction -f YOUR_FILE.xml'
#
#     HELP: 'python -m extractipedia.Extraction --help' to get more information.
#
#     =======================================================================

from xml.etree.ElementTree import iterparse
from time import perf_counter
import multiprocessing
from gc import collect
from extractipedia.utils import *
import datetime
import argparse
import queue

__version__ = '0.0.1'

class WikiText:

    def __init__(self, file, batch_size):
        self.file = file
        self.batch_size = batch_size

        self.current_batch = []
        self.current_batch_size = 0

        self.xml_parser = None

    @classmethod
    def strip_that(self, string):
        idx = string.rfind('}')
        if idx != -1:
            string = string[idx + 1:]
        return string  

    def __iter__(self):
        return self

    def __next__(self):

        articles_done = 0

        if self.xml_parser is None:
            self.xml_parser = iterparse(self.file, events=("start", "end"))

        for event, element in self.xml_parser:
            flag = self.strip_that(element.tag)

            if event == 'start':
                if flag == 'page':
                    title, redirect = '', ''
                    id = -100
                    revision = False
                    ns = 0
                elif flag == 'revision':
                    revision = True
            else:
                if flag == 'title':
                    title = element.text   
                elif flag == 'text':
                    text = element.text
                elif flag == 'id' and not revision:
                    id = int(element.text)
                elif flag == 'redirect':
                    redirect = element.attrib['title']
                elif flag == 'ns':
                    ns = int(element.text)
                elif flag == 'page':
                    try:
                        if ns == 0:
                            if len(redirect) == 0:
                                if 'may refer to' in text:
                                    pass
                                else:
                                    articles_done += 1
                                    TEMP_DICT = {title:[id, text]}
                                    self.current_batch.append(TEMP_DICT)
                                    self.current_batch_size += 1
                                    del TEMP_DICT 
                    except Exception as e:
                        print(e) 

                    title, text, redirect = '', '', ''
                    ns = -100     
                    element.clear() 

                if self.current_batch_size >= self.batch_size:
                    # Yield the current batch as a single string.
                    batch_to_yield = self.current_batch.copy()
                    self.current_batch.clear()
                    _ = collect()
                    self.current_batch_size = 0
                    return batch_to_yield, articles_done

        # Yield any remaining data in the last batch.
        if self.current_batch:
            batch_to_yield = self.current_batch.copy()
            self.current_batch.clear()
            _ = collect()
            self.current_batch_size = 0
            return batch_to_yield, articles_done
        
        raise StopIteration

SENTINEL = 0

def process_batch(batch, database_file, table_name, first_sentence, pages):

    """
    :params batch: chunks of data to be written to the database.
    :params database_file: name of the database.
    :params table_name: name of the table in the database.
    :params pages: page count for every process.
    """
    global SENTINEL
    if pages > SENTINEL:
        SENTINEL = pages
    logging.info(f'{SENTINEL} articles have been updated.')
    result = cleaning_text(batch, first_sentence)
    update_sqlite_table_with_dict(database_file, table_name, result)

class BatchProcessor:

    def __init__(self, max_queue_size, process_func, num_workers, database_file, table_name, first_sentence):
        self.max_queue_size = max_queue_size
        self.batch_queue = multiprocessing.Queue(maxsize=max_queue_size)
        self.process_func = process_func
        self.num_workers = num_workers
        self.stop_event = multiprocessing.Event()

        self.database_file = database_file
        self.table_name = table_name

        self.first_sentence = first_sentence

        self.page_count = 0

    def process_batches(self, data_generator):
        producer_process = multiprocessing.Process(target=self._produce_batches, args=(data_generator,))
        consumer_processes = [multiprocessing.Process(target=self._consume_batches, args=(TOTAL_ARTICLES,)) for _ in range(self.num_workers)]

        producer_process.start()
        for consumer_process in consumer_processes:
            consumer_process.start()

        producer_process.join()
        for consumer_process in consumer_processes:
            consumer_process.join()

    def _produce_batches(self, data_generator):
        for item, pages in data_generator:
            combined = (item, pages)
            self.batch_queue.put(combined)

        # Signal that there are no more items to be added to the queue.
        for _ in range(self.num_workers):
            self.batch_queue.put(None)

    def _consume_batches(self, TOTAL_ARTICLES):
        while not self.stop_event.is_set():
            try:
                result = self.batch_queue.get(timeout=1)
                batch, pages = result if isinstance (result, tuple) else (result, 0)
                if batch is None:
                    break  # Exit the loop when there is no more batches.

                TOTAL_ARTICLES.value += pages

                # Process the batch using the provided function.
                self.process_func(batch, self.database_file, self.table_name, self.first_sentence, TOTAL_ARTICLES.value)
            except queue.Empty:
                pass
            except Exception as e:
                print(f"Error: {e}")

    def stop(self):
        self.stop_event.set()

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Extractipedia')
    parser.add_argument("-f", "--file_name", type=str, help="XML File Name")
    parser.add_argument("-b", "--batch_size", type=int, default=2500, help="Batch Size (default=2500)")
    parser.add_argument("-d", "--database_file", type=str, default='new_database.db', help="Database Name (default=new_database.db)")
    parser.add_argument("-t", "--table_name", type=str, default='new_table', help="Table Name (default=new_table)")
    num_workers = multiprocessing.cpu_count() - 1
    max_workers = multiprocessing.cpu_count()
    parser.add_argument("-n", "--num_workers", type=int, default=num_workers, help=f"Number of Workers: (default={num_workers}, max={max_workers}). How many cores do you want to use? It is advisable that you should at least exclude 1 core in order to give your machine breathing room. Max is the core number your machine has. You can simply type the number of cores to be used.")
    parser.add_argument("-s", "--first_sentence", action="store_true", default=False, help="If you include [-s] while executing the script, it'll only write out the first sentences of the pages to the database. Useful for basic indexes. Reduce memory by more than 90 percent!")
    args = parser.parse_args()

    logging.info(f'File Name: {args.file_name}')
    logging.info(f'Batch Size: {args.batch_size}')
    logging.info(f'Database: {args.database_file}. Table: {args.table_name}')
    logging.info(f'Processors in Use: {args.num_workers}')
    logging.info(f'First Sentence? {args.first_sentence}')

    start_time = perf_counter()

    TOTAL_ARTICLES = multiprocessing.Value('i', 0)

    max_queue_size = 5

    batch_processor = BatchProcessor(max_queue_size, process_func=process_batch, num_workers=args.num_workers, database_file=args.database_file, table_name=args.table_name, first_sentence=args.first_sentence)
    generator = WikiText(args.file_name, args.batch_size)

    batch_processor.process_batches(generator)

    bed_time = perf_counter()

    time_difference = round(bed_time - start_time)
    formatted_time = str(datetime.timedelta(seconds=time_difference))

    logging.info(f'It is done in {formatted_time} (hh:mm:ss).')


from setuptools import find_packages, setup

setup(
    name = 'extractipedia',
    version = '0.0.1',
    description = 'A simple tool that extracts plain text Wikipedia Pages to SQLite database.',
    packages = find_packages(),
    long_description = open('README.md').read(),
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/sharkcommits/extractipedia',
    author = 'sharkcommits',
    author_email = 'sharkcommits@protonmail.com',
    license = 'GPLv3',
    classifiers = [
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    extras_require = {
        'dev': ['pytest>=7.0', 'twine>=4.0.2']    
    },
    python_requires = '>=3.10'
) 